#include "acl.h"
#include <arpa/inet.h>
#include <ifaddrs.h>
#include <pthread.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>


/*
*    Lookup telephone number AGID.
*    Required external program is curl
*
*    Author: R.J. van der Putten, Leiden, Holland, 
*    rob at sput dot nl.
* 
*    This program is free software; you can redistribute it and/or modify
*    it under the terms of the GNU General Public License version 3 as
*    published by the Free Software Foundation.
* 
*    This program is distributed in the hope that it will be useful,
*    but WITHOUT ANY WARRANTY; without even the implied warranty of
*    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*    GNU General Public License for more details.
*/


#pragma GCC diagnostic ignored "-Wpragmas"
#pragma GCC diagnostic ignored "-Wstringop-truncation"
#pragma GCC diagnostic ignored "-Wformat-truncation"


/*
*  Function list
*
*  Check ACL; chkacl()
*  Convert netmask; calcmask()
*  Clean up network address; cleanet()
*  Converts IPv4 to IPv6; fourtosix()
*  Automatic ACL; autoacl()
*  Parse ACL file; readacl()
*/


/*  Check ACL  */
int chkacl(struct in6_addr *remaddr)
{
    int i, j;

    if (numacl == 0 || acltab == NULL) {
	if(debflg != 0)
	    wrtolog("Debug: chkacl(): No ACL", NULL);
	return(0);
    }

    /*
    *  1234:5678:abcd	IP
    *  1234:5678:0000	Net
    *  ffff:ffff:0000	Mask
    */

    i = 0;
    while (i < numacl) {
	j = 0;
	while (j < 16) {
	    if ((remaddr->s6_addr[j] & acltab[i].mask[j]) != acltab[i].net[j]) {
		/*  No match  */
		break;
	    }
	    j++;
	    if (j == 16) {
		/*  All bytes match  */
		return(0);
	    }
	}
	i++;
    }

    /*  None match  */
    return(-1);
}

/*  Convert netmask  */
int calcmask(unsigned char *addr6, int mask)
{
    int i, j;

    i = 0;
    j = 0;

    if (mask == 0) {
	memset(addr6, 0, 16);
	return(0);
    } else {
	memset(addr6, 0xFF, 16);
    }

    if(mask == 128)
	return(0);

    /*
    *  0 1  2 3  4 5  6 7  8 9  A B  C D  E F
    *  ffff:ffff:ffff:0000:0000:0000:0000:0000
    *  0  10   20   30   40   50   60   70   80
    */

    i = 128;
    j = 15;
    while (i > mask) {
	addr6[j] = (addr6[j] << 1);
	i--;
	if(addr6[j] == 0)
	    j--;
    }

    return(0);
}

/*  Clean up network address  */
void cleanet(int net)
{
    int i;

    /*  Bitwise AND of network and netmask  */
    for (i = 0; i < 16; i++) {
	acltab[net].net[i] = (acltab[net].net[i] & acltab[net].mask[i]);
    }
}

/*  Converts IPv4 to IPv6  */
void fourtosix(unsigned char *addr6, unsigned int addr4, int ismask)
{
    if (ismask == 0) {
	/*  IPv4-mapped IPv6 => Insert ::ffff: bit  */
	memcpy(addr6, ip4adrprx, 12);
    } else {
	/*  Netmask => Insert ffff:ffff:ffff:ffff:ffff:ffff:   */
	memcpy(addr6, ip4mskprx, 12);
    }
    /*  Last four bytes (network byte order)  */
    memcpy(addr6 + 12, (unsigned char *) &addr4, 4);
}

/*  Automatic ACL  */
int autoacl(void)
{
    char *logstr;
    int aclcnt, entries, maxcnt, s;
    size_t size;
    struct ifaddrs *ifaddr;
    struct sockaddr_in *addr;
    struct sockaddr_in6 *addr6;
    struct in6_addr shwad6;
    unsigned int addr4, hstadr, mask4;

    aclcnt = 0;
    addr4  = 0;	/*  Network byte order  */
    hstadr = 0;	/*  Host byte order  */
    mask4  = 0;	/*  Network byte order  */
    aclcnt = 0;
    s      = 0;
    size   = 0;
    ifaddr = NULL;
    addr   = NULL;
    addr6  = NULL;
    ipstr[0] = 0;
    memset(&shwad6, 0, sizeof(struct in6_addr));

    logstr = glogs;
    logstr[0] = 0;

    /*  Init  */
    entries = 8;		/*  Nr of possible entries  */
    maxcnt = entries - 1;	/*  Minus margin  */
    size = (size_t) entries * sizeof(struct aclprps);
    acltab = malloc(size);
    if (acltab == NULL) {
	wrtolog("autoacl(): Cant allocate memory for ACL table", NULL);
        numacl = 0;
        return(-1);
    }
    memset(acltab, 0, size);
    snprintf(logstr, 255, \
      "autoacl(): Allocated %d bytes for acltab at %lx", \
      (int) size, (unsigned long) acltab);
    wrtolog(logstr, NULL);

    /*  int getifaddrs(struct ifaddrs **ifap);  */
    if (getifaddrs(&ifaddr) == -1) {
	free(acltab);
	numacl = 0;
	acltab = NULL;
	wrtolog("autoacl(): getifaddrs() failed", NULL);
	return(-1);
    }

    for (struct ifaddrs *ifa = ifaddr; ifa != NULL; ifa = ifa->ifa_next) {
	if (ifa->ifa_addr == NULL)
	    continue;

	if (ifa->ifa_addr->sa_family == AF_INET) {
	    /*  IPv4  */
	    addr = (struct sockaddr_in *) ifa->ifa_addr;
	    addr4 = addr->sin_addr.s_addr;
	    /*  Convert to host byte order  */
	    hstadr = ntohl(addr4);
	    addr = (struct sockaddr_in *) ifa->ifa_netmask;
	    mask4 = addr->sin_addr.s_addr;
	    if (hstadr == 0x7f000001) {
		/*  127.0.0.1 => force /32  */
		mask4 = 0xffffffff;
	    }
	    fourtosix(acltab[aclcnt].net, addr4, 0);
	    fourtosix(acltab[aclcnt].mask, mask4, 1);
	    if (hstadr != 0x7f000001) {
		/*  Cleanup network  */
		cleanet(aclcnt);
	    }
	    if (debflg != 0) {
		/*  Print net  */
		memcpy(shwad6.s6_addr, acltab[aclcnt].net, 16);
		if (inet_ntop(AF_INET6, &shwad6, ipstr, 47) != NULL) {
		    snprintf(logstr, 255, \
		      "Debug: autoacl(): net:  %s", ipstr);
		    wrtolog(logstr, NULL);
		}
		/*  Print mask  */
		memcpy(shwad6.s6_addr, acltab[aclcnt].mask, 16);
		if (inet_ntop(AF_INET6, &shwad6, ipstr, 47) != NULL) {
		    snprintf(logstr, 255, \
		      "Debug: autoacl(): mask: %s", ipstr);
		    wrtolog(logstr, NULL);
		}
            }
	    aclcnt++;
	} else if (ifa->ifa_addr->sa_family == AF_INET6) {
	    /*  IPv6  */
	    addr6 = (struct sockaddr_in6 *) ifa->ifa_addr;
	    if (addr6->sin6_addr.s6_addr[0] != 0xfe && \
	      addr6->sin6_addr.s6_addr[1] != 0x80) {
		/*  Not fe80::  */
		memcpy(acltab[aclcnt].net, addr6->sin6_addr.s6_addr, 16);
		addr6 = (struct sockaddr_in6 *) ifa->ifa_netmask;
		memcpy(acltab[aclcnt].mask, &(addr6->sin6_addr.s6_addr), 16);
		cleanet(aclcnt);
		if (debflg != 0) {
		    /*  Print net  */
		    memcpy(shwad6.s6_addr, acltab[aclcnt].net, 16);
		    if (inet_ntop(AF_INET6, &shwad6, ipstr, 47) != NULL) {
			snprintf(logstr, 255, \
			  "Debug: autoacl(): net:  %s", ipstr);
			wrtolog(logstr, NULL);
		    }
		    /*  Print mask  */
		    memcpy(shwad6.s6_addr, acltab[aclcnt].mask, 16);
		    if (inet_ntop(AF_INET6, &shwad6, ipstr, 47) != NULL) {
			snprintf(logstr, 255, \
			  "Debug: autoacl(): mask: %s", ipstr);
			wrtolog(logstr, NULL);
		    }
		}
		aclcnt++;
	    }
	}
	if (aclcnt >= maxcnt) {
	    /*  Increase size */
	    acltab = (struct aclprps *) realloc(acltab, (size_t) (size * 2));
	    if (acltab == NULL) {
		wrtolog("autoacl(): realloc() failed", NULL);
		numacl = 0;
		freeifaddrs(ifaddr);
		return(-1);
	    }
	    memset(acltab + entries, 0, size);
	    size = size * 2;
	    entries = entries * 2;
	    maxcnt = entries - 2;
	    snprintf(logstr, 255, \
	      "autoacl(): Allocated %d bytes for acltab at %lx", \
              (int) size, (unsigned long) acltab);
	    wrtolog(logstr, NULL);
	}
    }

    freeifaddrs(ifaddr);

    numacl = aclcnt;

    return(0);
}

/*  Parse ACL file  */
int readacl(void)
{
    char *aclnam, *line, *logstr;
    int aclcnt, i, j, isip6, lincnt, lins, mask, mskok;
    unsigned int net;
    size_t size;
    struct in_addr addr;
    struct in6_addr addr6, mask6;
    FILE *aclfil;

    aclcnt = 0;
    i = 0;
    j = 0;
    isip6;
    lincnt = 0;
    lins = 0;
    mask = 0;
    mskok = 0;
    net = 0;
    size = 0;
    aclfil = NULL;
    acltab = NULL;
    ipstr[0] = 0;
    memset(&addr,  0, sizeof(struct in_addr));
    memset(&addr6, 0, sizeof(struct in6_addr));
    memset(&mask6, 0, sizeof(struct in6_addr));

    line = genbuf;
    logstr = glogs;
    line[0] = 0;
    logstr[0] = 0;

    if (confdir[0] == 0)
	aclnam = ANLD_ACL_FILE;
    else
	aclnam = ANLD_ACL_FIL;

    if ((aclfil = fopen(aclnam, "r")) == NULL) {
	snprintf(logstr, 255, "readacl(): Cant open %s", aclnam);
	wrtolog(logstr, NULL);
	return(-1);
    }

    /*  Count lines  */
    while (fgets(line, ANLD_BUF_SIZ, aclfil)) {
	lincnt++;
    }

    lincnt += 3;	/*  Localhost plus margin  */
    size = (size_t) lincnt * sizeof(struct aclprps);
    acltab = malloc(size);
    if (acltab == NULL) {
	wrtsysl("readacl(): Cant allocate memory for ACL table", NULL, 1);
	fclose(aclfil);
	numacl = 0;
	return(-1);
    }
    if (debflg != 0) {
	snprintf(logstr, 255, \
	  "Debug: readacl(): allocated %d bytes for acltab at %lx", \
	  (int) size, (unsigned long) acltab);
	wrtolog(logstr, NULL);
    }
    memset(acltab, 0, size);

    /*  Add localhost  */
    memcpy(acltab[0].net,  localhost[0].net,  16);
    memcpy(acltab[0].mask, localhost[0].mask, 16);
    memcpy(acltab[1].net,  localhost[1].net,  16);
    memcpy(acltab[1].mask, localhost[1].mask, 16);
    aclcnt = 2;

    rewind(aclfil);

    /*  Read ACL file  */
    while (fgets(line, ANLD_BUF_SIZ, aclfil)) {
	if (line[0] == 10 || line[0] == 13 || line[0] == '#') {
	    /*  Skip  */
	    continue;
	}
	/*  Cleanup  */
	i = 0;
	j = 0;
	while (i < ANLD_BUF_SIZ) {
	    if (line[i] == 10 || line[i] == 13 || \
	      line[i] == ' ' || line[i] == '#') {
		line[i] = 0;
		break;
	    } else if (line[i] == '.') {
		isip6 = 0;
	    } else if (line[i] == ':') {
		isip6 = 1;
	    } else if (line[i] == '/') {
		line[i] = 0;
		j = i + 1;
	    }
	    i++;
	}
	mskok = 0;
	if (isip6 == 0) {
	    if (j == 0) {
		mask = 32;
	    } else {
		mask = atoi(line + j);
		if (mask < 0 || mask > 32) {
		    wrtolog("readacl(): Invalid mask", NULL);
		    mskok = -1;
		}
	    }
	} else {
	    if (j == 0) {
		mask = 128;
	    } else {
		mask = atoi(line + j);
		if (mask < 0 || mask > 128) {
		    wrtolog("readacl(): Invalid mask", NULL);
		    mskok = -1;
		}
	    }
	}
	if (mskok == 0) {
	    /*  Netmask is OK  */
	    if (isip6 == 0) {
		if (inet_pton(AF_INET, line, &addr) == 1)
		    /*  Valid IPv4 address */
		    net = addr.s_addr;
		    fourtosix(acltab[aclcnt].net, net, 0);
		    calcmask(acltab[aclcnt].mask, mask + 96);
		    /*  Cleanup network  */
		    cleanet(aclcnt);
		    if (debflg != 0) {
			memcpy(addr6.s6_addr, acltab[aclcnt].net, 16);
			if (inet_ntop(AF_INET6, &addr6, ipstr, 47) != NULL) {
			    snprintf(logstr, 255, \
			      "Debug: readacl(): net:  %s", ipstr);
			    wrtolog(logstr, NULL);
			}
			memcpy(addr6.s6_addr, acltab[aclcnt].mask, 16);
			if (inet_ntop(AF_INET6, &addr6, ipstr, 47) != NULL) {
			    snprintf(logstr, 255, \
			      "Debug: readacl(): mask: %s", ipstr);
			    wrtolog(logstr, NULL);
			}
		    }
		    aclcnt++;
	    } else {
		if (inet_pton(AF_INET6, line, &addr6) == 1) {
		    /*  Valid IPv6 address */
		    memcpy(acltab[aclcnt].net, addr6.s6_addr, 16);
		    calcmask(acltab[aclcnt].mask, mask);
		    /*  Cleanup network  */
		    cleanet(aclcnt);
		    if (debflg != 0) {
			/*  Print net  */
			memcpy(addr6.s6_addr, acltab[aclcnt].net, 16);
			if (inet_ntop(AF_INET6, &addr6, ipstr, 47) != NULL) {
			    snprintf(logstr, 255, \
			      "Debug: readacl(): net:  %s", ipstr);
			    wrtolog(logstr, NULL);
			}
			/*  Print mask  */
			memcpy(addr6.s6_addr, acltab[aclcnt].mask, 16);
			if (inet_ntop(AF_INET6, &addr6, ipstr, 47) != NULL) {
			    snprintf(logstr, 255, \
			      "Debug: readacl(): mask: %s", ipstr);
			    wrtolog(logstr, NULL);
			}
		    }
		    aclcnt++;
		}
	    }
	}
	lincnt--;
	if (lincnt <= 0) {
	    wrtolog("readacl(): ACL File changed", NULL);
	    free(acltab);
	    acltab = NULL;
	    aclcnt = 0;
	    break;
	}
    }

    fclose(aclfil);

    numacl = aclcnt;

    return(0);
}

