#!/bin/bash

CURDIR=""
RUN=""

# Destination directories
SBINDIR="/usr/local/sbin"
DOCDIR="/usr/local/share/doc/namelookup"
HOMEDIR="/var/local/lib/namelookup"
LOGDIR="/var/local/log/namelookup"
MAN5DIR="/usr/local/share/man/man5"
MAN8DIR="/usr/local/share/man/man8"
UTILDIR="/usr/local/share/phonebook/utils"

# User
USER="namelookup"


# Copy and strip if newer
copy_bin() {
	# Original Path/File from tar
	ORIG="${1}"
	# Copy to this dir
	DIR="${2}"
	# Just the filename
	FILE=$( basename "${ORIG}" )
	if ! [ -f "${DIR}/${FILE}" ] || [ "${ORIG}" -nt "${DIR}/${FILE}" ]
	then
		cp --preserve=mode,timestamp "${ORIG}" "${DIR}/"
		strip "${DIR}/${FILE}"
	fi
}

# Copy and gzip man page if newer
copy_man() {
	# Original Path/File from tar
	ORIG="${1}"
	# Copy to this dir
	DIR="${2}"
	# Just the filename
	FILE=$( basename "${ORIG}" )
	if ! [ -f "${DIR}/${FILE}.gz" ] || [ "${ORIG}" -nt "${DIR}/${FILE}.gz" ]
	then
		cp --preserve=timestamp "${ORIG}" "${DIR}/"
		gzip -f "${DIR}/${FILE}"
	fi
}


# Check UID
if [ "${UID}" -ne 0 ]
then
	echo "You need to be root to install."
	echo "Aborting install."
	exit 1
fi

# Install binary
if [ ! -d "${SBINDIR}" ]
then
	echo "Creating dir ${SBINDIR}."
	mkdir -p "${SBINDIR}"
fi
echo "Copying binaries"
copy_bin namelookup.agid "${SBINDIR}"

# Install util
if [ ! -d "${UTILDIR}" ]
then
	echo "Creating dir ${UTILDIR}"
	mkdir -p "${UTILDIR}"
fi
echo "Copying utils"
copy_bin block2wide "${UTILDIR}"

# Install doc
if [ ! -d "${DOCDIR}" ]
then
	echo "Creating dir ${DOCDIR}."
	mkdir -p "${DOCDIR}"
fi
echo "Copying doc"
cp -ru --preserve=timestamp doc/* "${DOCDIR}/"

# Install man pages
if [ ! -d "${MAN5DIR}" ]
then
	echo "Creating dir ${MAN5DIR}."
	mkdir -p "${MAN5DIR}"
fi
if [ ! -d "${MAN8DIR}" ]
then
	echo "Creating dir ${MAN8DIR}."
	mkdir -p "${MAN8DIR}"
fi
echo "Copying man pages."
copy_man man/namelookup.acl.5  "${MAN5DIR}"
copy_man man/namelookup.agid.8 "${MAN8DIR}"

# Group and user
if ( getent group "${USER}" > /dev/null )
then
	echo "Group ${USER} exists"
else
	echo "Adding group ${USER}."
	addgroup --system "${USER}"
fi
if ( getent passwd "${USER}" > /dev/null )
then
	echo "User ${USER} exists"
else
	echo "Adding user ${USER}."
	adduser --system --home "${HOMEDIR}" \
	    --disabled-password --ingroup "${USER}" "${USER}"
fi

# Log dir
if [ -d "${LOGDIR}" ]
then
	echo "Dir ${LOGDIR} exists."
else
	echo "Creating dir ${LOGDIR}."
	mkdir -p "${LOGDIR}"
	chmod 775 "${LOGDIR}"
	chown "${USER}":"${USER}" "${LOGDIR}"
fi

# Log file rotation.
if [ -d "/etc/logrotate.d/" ]
then
	echo "Copying log rotate file."
	echo "Note: You may want to edit /etc/logrotate.d/namelookup."
	cp -u --preserve=timestamp doc/etc/logrotate.d/namelookup \
	    /etc/logrotate.d/
else
	echo "/etc/logrotate.d/ not found"
fi

# SysV / Insserv install
if [ -d "/etc/default/" ]
then
	cp -u --preserve=timestamp doc/etc/default/namelookup-agid \
	    /etc/default/
else
	"/etc/default/ not found"
fi
if [ -d "/etc/init.d/" ]
then
	cp -u --preserve=timestamp doc/etc/init.d/namelookup-agid \
	    /etc/init.d/
else
	"/etc/init.d/ not found"
fi

# Systemd install
if [ -d "/lib/systemd/system/" ]
then 
	cp -u --preserve=timestamp doc/systemd/namelookup-agid.service \
	    /lib/systemd/system/
else
	echo "/lib/systemd/system/ not found"
fi

# Start on boot
RUN="N"
if [ -x /usr/bin/systemctl ]
then
	echo "Making Systemd symlink"
	systemctl enable namelookup-agid.service
	if [ -x /usr/sbin/update-rc.d ]
	then
		echo "Do you want SysV enabled as well?"
		echo "Type 'y' <Enter> for yes."
		read -r RUN
		if [ "${RUN}" = "Y" ] || [ "${RUN}" = "y" ] || \
		    [ "${RUN}" = "YES" ] || [ "${RUN}" = "Yes" ] || \
		    [ "${RUN}" = "yes" ]
		then
			update-rc.d namelookup-agid defaults
		fi
	fi
elif [ -x /usr/sbin/update-rc.d ] || [ -x /sbin/update-rc.d ]
then
	echo "Making SysV RC symlinks"
	update-rc.d namelookup-agid defaults
	if [ -d "/etc/systemd/system/multi-user.target.wants/" ] && \
	    [ -d "/lib/systemd/system/" ]
	then
		echo "Do you want Systemd symlinks as well?"
		echo "Type 'y' <Enter> for yes."
		read -r RUN
		if [ "${RUN}" = "Y" ] || [ "${RUN}" = "y" ] || \
		    [ "${RUN}" = "YES" ] || [ "${RUN}" = "Yes" ] || \
		    [ "${RUN}" = "yes" ]
		then
			CURDIR=$( pwd )
			cd "/etc/systemd/system/multi-user.target.wants/"
			ln -s /lib/systemd/system/namelookup-agid.service \
			    namelookup-agid.service
			cd "${CURDIR}"
		fi
	fi
else
	echo "Both update-rc.d and systemctl where not found"
	echo "namelookup.agid will not automatically start on boot"
fi

# Ncurses config tool
RUN="N"
if [ -x "/usr/local/bin/agiconf" ]
then
	if ( /usr/local/bin/agiconf -v | \
	    grep "With name length 71 bytes" > /dev/null )
	then
		echo ""
		echo "Do you want agiconf to configure namelookup-agi now?"
		echo "Type 'y' <Enter> for yes."
		echo "Note: You need to select the compact file option."
		echo "Note: You can run agiconf later if you like."
		read -r RUN
		if [ "${RUN}" = "Y" ] || [ "${RUN}" = "y" ] || \
		    [ "${RUN}" = "YES" ] || [ "${RUN}" = "Yes" ] || \
		    [ "${RUN}" = "yes" ]
		then
			agiconf -c -g "${USER}"
			#agiconf -ac -g "${USER}"
		else
			echo "Not running agiconf"
		fi
	else
		echo "agiconf needs to be compiled with the WIDE option."
	fi
else
	echo "agiconf not foud"
fi

RUN="N"
echo ""
echo "You may want to edit /etc/default/namelookup-agid before starting this"
echo "daemon;"
echo "If you are already using FastAGI / port 4573, you need to set"
echo "an other port in this file. E.G.:"
echo "OPTIONS=\"-p 4574 -u namelookup\"."
echo ""
echo "Do you want to start this daemon now?"
echo "Type 'y' <Enter> for yes."
echo "Note: You can start this daemon later if you like"
read -r RUN
if [ "${RUN}" = "Y" ] || [ "${RUN}" = "y" ] || \
     [ "${RUN}" = "YES" ] || [ "${RUN}" = "Yes" ] || [ "${RUN}" = "yes" ]
then
	if [ -d /run/systemd/system ]
	then
		systemctl start namelookup-agid
	else
		/etc/init.d/namelookup-agid start
	fi
else
	echo "Not starting daemon"
fi

