#include <arpa/inet.h>
#include <errno.h>
#include <dirent.h>
#include <fcntl.h>
#include <grp.h>
#include <locale.h>
#include <pwd.h>
#include <pthread.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/mman.h>
#include <sys/select.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <sys/types.h>
#include <syslog.h>
#include <unistd.h>
#include <wctype.h>
#include "main.h"
#include "version.h"


/*
*    Lookup telephone number AGID.
*    Required external program is curl
*
*    Author: R.J. van der Putten, Leiden, Holland, 
*    rob at sput dot nl.
* 
*    This program is free software; you can redistribute it and/or modify
*    it under the terms of the GNU General Public License version 3 as
*    published by the Free Software Foundation.
* 
*    This program is distributed in the hope that it will be useful,
*    but WITHOUT ANY WARRANTY; without even the implied warranty of
*    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*    GNU General Public License for more details.
*/


#pragma GCC diagnostic ignored "-Wpragmas"
#pragma GCC diagnostic ignored "-Wstringop-truncation"
#pragma GCC diagnostic ignored "-Wformat-truncation"


/*
*  This will actually demonize
*  Comment out for tests
*/
#define AGID_DAEMON 1


/*  General  */
int contprog;		/*  Continue program  */
int debflg;		/*  Debug on if not zero  */
int lsfd;		/*  Listen Socket FD  */
char confdir[256];	/*  Alternate conf dir  */
char genbuf[4096];	/*  General purpose buf. Not used by child threads  */
char glogs[256];	/*  Global string for log. Not used by child threads  */
char ipstr[48];		/*  Text version of IP address. Not used by kids */

/*  ACL  */
int aclflg;
int numacl;
struct aclprps *acltab;

/*  Threads  */
struct thrdprps *thrdtab;
int numthrds;

/*  Files  */
FILE *flog;

/*  
*  Lookup tables
*/

/*  Block list  */
char  *blkdb;
int    blkent;
size_t blksiz;
/*  Contacts, area codes and countries  */
struct phndb numbers;
struct phndb areas;
struct phndb regions;
/*  Weblookup  */
int disweb;
struct urlprps lkpsite;

/*
*  Function list
*
*  Print help; hlp()
*  Print version; prtvers()
*  Write to syslog; wrtsysl()
*  Write to log; wrtolog()
*  Free and close stuff; quit()
*  Rotate log; rotlog()
*  Stop program; stop()
*
*  Load database; loadb()
*  Load databases; loadbs()
*  Check for quotes; chkquot()
*  Read web config file; readwebcnf()
*  Initialize thread table; inithrdtab()
*  Init program; init()
*
*  Convert name to lower case; namtolow()
*  Names compare; compnam()
*  Lookup in blacklist; blklookup()
*  Phone numbers index compare; compidx()
*  Lookup name or location in database; phndblookup()
*
*  Find name in buf; findname()
*  Lookup name on web; weblookup()
*
*  Main: main()
*/

/*
*  Misc functions
*/

/*  Help  */
void hlp(void)
{
    fprintf(stderr, "Asterisk name and location lookup AGI daemon\n");
    fprintf(stderr, "Usage: namelookup.agid [-a] [-c confdir] [-d] [-h]\n");
    fprintf(stderr, "\t[-n threads] [-p port] [-u user] [-v]\n\n");
    fprintf(stderr, "-a: Automatic ACL.\n");
    fprintf(stderr, "-c: Alternate config dir.\n");
    fprintf(stderr, "-d: Debug on.\n");
    fprintf(stderr, "-h: Print this Help and exit.\n");
    fprintf(stderr, "-n: Number of threads. 2 - 1024. Default: 16\n");
    fprintf(stderr, "-p: Port number. 1024 - 65535. Default: 4573\n");
    fprintf(stderr, "-u: User. This also sets the group.\n");
    fprintf(stderr, "-v: Print Version and exit.\n\n");
}

/*  Print version  */
void prtvers(void)
{
    printf("Version: %s\n", version);
}

/*  Write to syslog  */
void wrtsysl(char *str1, char *str2, int mode)
{
    int prio;

    if(mode == 0)
	prio = LOG_INFO;
    else
	prio = LOG_ERR;

    openlog("namelookup", LOG_PID, LOG_USER);
    if (str2 == NULL)
	syslog(prio, "%s", str1);
    else
	syslog(prio, "%s: %s", str1, str2);
    closelog();
}

/*  Write to log  */
int wrtolog(char *s1, char *s2)
{
    char datstr[32], zonstr[8];
    struct timeval logtime;
    struct tm *loc;

    loc = NULL;

    if (gettimeofday(&logtime, NULL) != 0) {
	wrtsysl("wrtolog(): gettimeofday() failed", strerror(errno), 1);
	return(-1);
    }

    loc = localtime(&logtime.tv_sec);
    /*  YYYY-MM-DD hh:mm:ss.uuuuuu timezone  */
    strftime(datstr, 31, "%F %T", loc);
    strftime(zonstr, 7, "%z", loc);
    if (s2 == NULL) {
	fprintf(flog, "%s.%06ld %s %s\n", datstr, \
	  logtime.tv_usec, zonstr, s1);
    } else {
	fprintf(flog, "%s.%06ld %s %s: %s\n", \
	  datstr, logtime.tv_usec, zonstr, s1, s2);
    }
    fflush(flog);

    return(0);
}

/*  Free and close stuff  */
void quit(void)
{
    if (lsfd > -1) {
	close(lsfd);
	lsfd = -1;
    }
    if (thrdtab != NULL) {
	free(thrdtab);
	thrdtab = NULL;
    }
    if (acltab != NULL) {
	free(acltab);
	acltab = NULL;
    }

    if (regions.lst != NULL) {
	munmap(regions.lst, regions.lstsiz);
	regions.lst = NULL;
    }
    if (regions.idx != NULL) {
	munmap(regions.idx, regions.idxsiz);
	regions.idx = NULL;
    }
    if (areas.lst != NULL) {
	munmap(areas.lst, areas.lstsiz);
	areas.lst = NULL;
    }
    if (areas.idx != NULL) {
	munmap(areas.idx, areas.idxsiz);
	areas.idx = NULL;
    }
    if (numbers.lst != NULL) {
	munmap(numbers.lst, numbers.lstsiz);
	numbers.lst = NULL;
    }
    if (numbers.idx != NULL) {
	munmap(numbers.idx, numbers.idxsiz);
	numbers.idx = NULL;
    }
    if (blkdb != NULL) {
	munmap(blkdb, blksiz);
	blkdb = NULL;
    }

    if (flog != NULL) {
	fclose(flog);
	flog = NULL;
    }
}

/*  Rotate log  */
void rotlog(int sig)
{
    char *lognam; 

    if (flog != NULL) {
	fclose(flog);

	if (debflg == 0) {
	    lognam = ANLD_LOG_FILE;
	} else {
	    /*  Log file in default dir  */
	    lognam = ANLD_LOG_FIL;
	}
	if ((flog = fopen(lognam, "a")) == NULL) {
	    wrtsysl("rotlog(): Cant open", lognam, 1);
	    contprog = 0;
	}
    }
}

/*  Stop program  */
void stop(int sig)
{
    contprog = 0;
}

/*
*  Init and load databases
*/

/*  Load database  */
void * loadb(char *filnam, size_t *size)
{
    int dbfd;
    size_t dbsiz;
    void *dbaddr;
    char *logstr;
    struct stat sb;

    dbfd = 0;
    dbsiz = 0;
    dbaddr = NULL;
    memset(&sb, 0, sizeof(sb));

    logstr = glogs;
    logstr[0] = 0;

    /*  Open file  */
    if ((dbfd = open(filnam, O_RDONLY)) < 0) {
	wrtsysl("loadb(): Cant open file", filnam, 1);
	*size = 0;
	return(NULL);
    }

    if (fstat(dbfd, &sb) != 0) {
	wrtsysl("loadb(): fstat() failed", NULL, 1);
	close(dbfd);
	*size = 0;
	return(NULL);
    }
    dbsiz = (size_t) sb.st_size;

    dbaddr = mmap(NULL, dbsiz, PROT_READ, MAP_PRIVATE, dbfd, 0);
    close(dbfd);
    if (dbaddr == MAP_FAILED) {
	wrtsysl("loadb(): mmap() failed", NULL, 1);
	*size = 0;
	return(NULL);
    }

    *size = dbsiz;

    if (debflg != 0) {
	snprintf(logstr, 255, \
	  "Debug: loadb(): loaded %s of %d bytes at %lx", \
	  filnam, (int) *size, (unsigned long) dbaddr);
	wrtolog(logstr, NULL);
    }

    return(dbaddr);
}

/*  Load databases  */
int loadbs(void)
{
    char *filnam, *logstr;

    filnam = NULL;

    logstr = glogs;
    logstr[0] = 0;

    /*  Blocklist  */
    if(confdir[0] == 0)
	filnam = ANLD_BLOCK_FILE;
    else
	filnam = ANLD_BLCK_FIL;
    blkdb = (char *) loadb(filnam, &blksiz);
    if (blkdb == NULL) {
	wrtsysl("loadbs: Cant load", filnam, 0);
	blkent = 0;
    } else {
	blkent = (int) blksiz / ANLD_NAME_SIZ;
	if (debflg != 0) {
	    snprintf(logstr, 255, "Debug: loadbs(): %d entries in blocklist", \
	      blkent);
	    wrtolog(logstr, NULL);
	}
    }

    /*  Contact index  */
    if(confdir[0] == 0)
	filnam = ANLD_NUMBERS_INDEX;
    else
	filnam = ANLD_NUMS_IDX;
    numbers.idx = (struct phidx *) loadb(filnam, &(numbers.idxsiz));
    if (numbers.idx == NULL) {
	wrtsysl("loadbs: Cant load", filnam, 0);
	numbers.ent = 0;
    } else {
	/*  Number of entries  */
	numbers.ent = (int) (numbers.idxsiz / sizeof(struct phidx));
	if (debflg != 0) {
	    snprintf(logstr, 255, "Debug: loadbs(): %d entries in numbers", \
	      numbers.ent);
	    wrtolog(logstr, NULL);
	}
    }
    /*  Contacts  */
    if (numbers.idx != NULL) {
	/*  numbers.idx exists  */
	if(confdir[0] == 0)
	    filnam = ANLD_NUMBERS_LIST;
	else
	    filnam = ANLD_NUMS_LST;
	numbers.lst = (char *) loadb(filnam, &(numbers.lstsiz));
	if (numbers.lst == NULL) {
	    wrtsysl("loadbs(): Cant load", filnam, 1);
	    /*  Free numbers.idx  */
	    munmap(numbers.idx, numbers.idxsiz);
	    numbers.idx = NULL;
	    numbers.ent = 0;
	    wrtsysl("loadbs(): numbers.idx cleared", NULL, 1);
	}
    }

    /*  Area code index  */
    if(confdir[0] == 0)
	filnam = ANLD_AREAS_INDEX;
    else
	filnam = ANLD_AREA_IDX;
    areas.idx = (struct phidx *) loadb(filnam, &(areas.idxsiz));
    if (areas.idx == NULL) {
	wrtsysl("loadbs: Cant load", filnam, 0);
	areas.ent = 0;
    } else {
	/*  Number of entries  */
	areas.ent = (int) (areas.idxsiz / sizeof(struct phidx));
	if (debflg != 0) {
	    snprintf(logstr, 255, "Debug: loadbs(): %d entries in areas", \
	      areas.ent);
	    wrtolog(logstr, NULL);
	}
    }
    /*  Area codes  */
    if (areas.idx != NULL) {
	/*  areas.idx exists  */
	if(confdir[0] == 0)
	    filnam = ANLD_AREAS_LIST;
	else
	    filnam = ANLD_AREA_LST;
	areas.lst = (char *) loadb(filnam, &(areas.lstsiz));
	if (areas.lst == NULL) {
	    wrtsysl("loadbs(): Cant load", filnam, 1);
	    /*  Free areas.idx  */
	    munmap(areas.idx, areas.idxsiz);
	    areas.idx = NULL;
	    areas.ent = 0;
	    wrtsysl("loadbs(): areas.idx cleared", NULL, 1);
	}
    }

    /*  Country code index */
    if(confdir[0] == 0)
	filnam = ANLD_REGIONS_INDEX;
    else
	filnam = ANLD_REGS_IDX;
    regions.idx = (struct phidx *) loadb(filnam, &(regions.idxsiz));
    if (regions.idx == NULL) {
	wrtsysl("loadbs: Cant load", filnam, 0);
	regions.ent = 0;
    } else {
	/*  Number of entries  */
	regions.ent = (int) (regions.idxsiz / sizeof(struct phidx));
	if (debflg != 0) {
	    snprintf(logstr, 255, "Debug: loadbs(): %d entries in regions", \
	      regions.ent);
	    wrtolog(logstr, NULL);
	}
    }
    /*  Country codes */
    if (regions.idx != NULL) {
	if(confdir[0] == 0)
	    filnam = ANLD_REGIONS_LIST;
	else
	    filnam = ANLD_REGS_LST;
	regions.lst = (char *) loadb(filnam, &(regions.lstsiz));
	if (regions.lst == NULL) {
	    wrtsysl("loadbs(): Cant load", filnam, 1);
	    /*  Free regions.idx  */
	    munmap(regions.idx, regions.idxsiz);
	    regions.idx = NULL;
	    regions.ent = 0;
	    wrtsysl("loadbs(): regions.idx cleared", NULL, 1);
	}
    }

    if (debflg != 0) {
	if (blkdb == NULL)
	    wrtolog("Debug: loadbs(): No blocklist DB", NULL);
	if (numbers.idx == NULL)
	    wrtolog("Debug: loadbs(): No numbers DB", NULL);
	if (areas.idx == NULL)
	    wrtolog("Debug: loadbs(): No areas DB", NULL);
	if (regions.idx == NULL)
	    wrtolog("Debug: loadbs(): No regions DB", NULL);
    }

    return(0);
}

/*  Check for quotes  */
int chkquot(char *line, int ofs, int len)
{
    if ((line[ofs] == '\"' && line[len - 1] == '\"') || \
      (line[ofs] == '\'' && line[len - 1] == '\'')) {
	line[len - 1] = 0;
	return(ofs + 1);
    } else {
	return(ofs);
    }
}

/*  Read web config file  */
int readwebcnf(void)
{
    int i, len, ofs;
    char *libnam, *line, *logstr;
    FILE *libfil;

    i = 0;
    len = 0;
    ofs = 0;
    libfil = NULL;

    line = genbuf;
    logstr = glogs;
    line[0] = 0;
    logstr[0] = 0;

    if (confdir[0] == 0) {
	libnam = ANLD_WEBCNF_FILE;
    } else {
	libnam = ANLD_WEBC_FIL;
    }

    if ((libfil = fopen(libnam, "r")) == NULL) {
	wrtsysl("readwebcnf: Cant open", libnam, 0);
	return(-1);
    }

    while (fgets(line, 4096, libfil)) {
	if (line[0] == 10 || line[0] == 13 || line[0] == '#')
	    continue;
	/*  Cleanup  */
	i = 1;
	while (line[i] != 0) {
	    if(line[i] == 10 || line[i] == 13) {
		line[i] = 0;
		len = i;
		break;
	    }
	    i++;
	}
	if (strncmp(line, "preget=", 7) == 0) {
	    /*  Check for quotes  */
	    ofs = chkquot(line, 7, len);
	    strncpy(lkpsite.preget, line + ofs, 255);
	    lkpsite.preget[255] = 0;
	} else if (strncmp(line, "pstget=", 7) == 0) {
	    ofs = chkquot(line, 7, len);
	    strncpy(lkpsite.pstget, line + ofs, 255);
	    lkpsite.pstget[255] = 0;
	} else if (strncmp(line, "prepost=", 8) == 0) {
	    ofs = chkquot(line, 8, len);
	    strncpy(lkpsite.prepost, line + ofs, 255);
	    lkpsite.prepost[255] = 0;
	} else if (strncmp(line, "pstpost=", 8) == 0) {
	    ofs = chkquot(line, 8, len);
	    strncpy(lkpsite.pstpost, line + ofs, 255);
	    lkpsite.pstpost[255] = 0;
	} else if (strncmp(line, "prename=", 8) == 0) {
	    ofs = chkquot(line, 8, len);
	    strncpy(lkpsite.prename, line + ofs, 255);
	    lkpsite.prename[255] = 0;
	} else if (strncmp(line, "pstname=", 8) == 0) {
	    ofs = chkquot(line, 8, len);
	    strncpy(lkpsite.pstname, line + ofs, 255);
	    lkpsite.pstname[255] = 0;
	} else if (strncmp(line, "atlin=", 6) == 0) {
	    lkpsite.atlin = atoi(line + 6);
	    if(lkpsite.atlin < 0)
		lkpsite.atlin = 0;
	} else if (strncmp(line, "atcol=", 6) == 0) {
	    lkpsite.atcol = atoi(line + 6);
	    if(lkpsite.atcol < 0)
		lkpsite.atcol = 0;
	}
    }

    fclose(libfil);

    lkpsite.prenmlen = strlen(lkpsite.prename);
    lkpsite.pstnmlen = strlen(lkpsite.pstname);

    if (debflg != 0) {
	if (lkpsite.preget[0] != 0)
	    wrtolog("Debug: readwebcnf(): preget", lkpsite.preget);
	if (lkpsite.pstget[0] != 0)
	    wrtolog("Debug: readwebcnf(): pstget", lkpsite.pstget);
	if (lkpsite.prepost[0] != 0)
	    wrtolog("Debug: readwebcnf(): prepost", lkpsite.prepost);
	if (lkpsite.pstpost[0] != 0)
	    wrtolog("Debug: readwebcnf(): pstpost", lkpsite.pstpost);
	if (lkpsite.prename[0] != 0)
	    wrtolog("Debug: readwebcnf(): prename", lkpsite.prename);
	if (lkpsite.pstname[0] != 0)
	    wrtolog("Debug: readwebcnf(): pstname", lkpsite.pstname);
	if (lkpsite.atlin != 0) {
	    snprintf(logstr, 255, "Debug: readwebcnf(): atlin: %d", \
	      lkpsite.atlin);
	    wrtolog(logstr, NULL);
	}
	if (lkpsite.atcol != 0) {
	    snprintf(logstr, 255, "Debug: readwebcnf(): atcol: %d", \
	      lkpsite.atcol);
	    wrtolog(logstr, NULL);
	}
    }

    if (lkpsite.prenmlen == 0 && lkpsite.atcol != 0) {
	/*
	*  There is no pre-name string.
	*  And findname() looks _AFTER_ the specified location.
	*  Decrement the colomn by one to compensate.
	*/
        (lkpsite.atcol)--;
    }

    return(0);
}

/*  Initialize thread table  */
int inithrdtab(void)
{
    int i;
    size_t size;
    char *logstr;

    logstr = glogs;
    logstr[0] = 0;

    size = (size_t) numthrds * sizeof(struct thrdprps);
    thrdtab = malloc(size);
    if (thrdtab == NULL)
	return(-1);
    memset(thrdtab, 0, size);

    /*  Init  */
    for (i = 0; i < numthrds; i++) {
	thrdtab[i].id  = i;
	thrdtab[i].sfd = -1;
    }

    if (debflg != 0) {
	snprintf(logstr, 255, 
	  "Debug: inithrdtab(): allocated %d bytes for %d threads at %lx", \
	  (int) size, numthrds, (unsigned long) thrdtab);
	wrtolog(logstr, NULL);
    }

    return(0);
}

/*  Init program  */
int init(void)
{
    if (loadbs() != 0) {
	quit();
	return(-1);
    }

    if (disweb == 0 && readwebcnf() != 0) {
	disweb = 1;
	wrtsysl("init: Web lookups disabled", NULL, 0);
    }

    /*  ACL  */
    if (aclflg == 0) {
	/*  Read ACL file if any  */
	readacl();
    } else {
	/*  Automatic ACL  */
	autoacl();
    }

    if (inithrdtab() != 0) {
	quit();
	return(-1);
    }

    return(0);
}

/*
*  Proces requests
*/

/*  Database stuff  */

/*  Convert name to lower case  */
int namtolow(char *name, char *lnam, wchar_t *wnam)
{
    int i; 
    size_t wlen;

    i = 0;
    wlen = 0;

    if(debflg != 0)
	wrtolog("Debug: namtolow() name", name);

    /*  Convert to wide characters  */
    if ((wlen = mbstowcs(wnam, name, ANLD_NAME_MAX)) < 0) {
	wrtsysl("namtolow(): mbstowcs() failed", NULL, 1);
	return(-1);
    }
    /*  Convert to lower case  */
    for(i = 0; i < (int) wlen; i++)
        wnam[i] = towlower(wnam[i]);
    /*  Convert to UTF-8  */
    if (wcstombs(lnam, wnam, ANLD_NAME_MAX) < 0) {
	wrtsysl("namtolow(): mbstowcs() failed", NULL, 1);
	return(-1);
    }

    if(debflg != 0)
	wrtolog("Debug: namtolow() converted to", lnam);

    return(0);
}

/*  Names compare  */
static int compnam(const void *n1, const void *n2)
{       
    char *name1, *name2;
            
    name1 = (char *) n1;
    name2 = (char *) n2;

    return(strcmp(name1, name2));
}

/*  Lookup in blacklist  */
int blklookup(char *name)
{
    char *res;

    res = NULL;

    if (blkdb == NULL) {
	if(debflg != 0)
	    wrtolog("Debug: blklookup(): No blocklist", NULL);
	return(0);
    }

    res = bsearch(name, blkdb, blkent, ANLD_NAME_SIZ, compnam);
    if (res != NULL) {
	wrtsysl("blklookup: Invalid", res, 0);
	return(1);
    }

    return(0);
}

/*  Phone numbers index compare  */
static int compidx(const void *n1, const void *n2)
{           
    int len;
    struct phidx *pho1, *pho2;
    
    pho1 = (struct phidx *) n1;    
    pho2 = (struct phidx *) n2;
    
    /*  One of these may contain the length  */
    if(pho1->len == 0)
	len = pho2->len;
    else
	len = pho1->len;
     
    if (len == 0) {
	/*  Neither contains a length; Exact match  */
	return(strcmp(pho1->num, pho2->num));
    } else {
	/*  Match the first bit  */
	return(strncmp(pho1->num, pho2->num, len));
    }
}

/*  Lookup name or location in database  */
int phndblookup(char *num, char *name, struct phndb db) 
{
    struct phidx key, *res;

    res = NULL;
    memset(&key, 0, sizeof(struct phidx));

    if (db.idx == NULL || db.lst == NULL) {
	if(debflg != 0)
	    wrtolog("Debug: phndblookup(): Database missing", NULL);
	return(0);
    }

    strncpy(key.num, num, ANLD_NUM_MAX);
    key.num[ANLD_NUM_MAX] = 0;

    res = bsearch(&key, db.idx, db.ent, sizeof(struct phidx), compidx);
    if (res != NULL) {
	if (res->namlen < 1 || res->namlen > ANLD_NAME_MAX || \
	  res->offs >= db.lstsiz) {
	    wrtsysl("phndblookup(): Error in index file", NULL, 1);
	} else {
	    strncpy(name, db.lst + res->offs, ANLD_NAME_MAX);
	    name[ANLD_NAME_MAX] = 0;
	}
    }

    return(0);
}

/*  Web lookup stuff  */

/*  Find name in buf  */
int findname(int id, int len)
{
    char *buf, ch, *name, *prenam, *pstnam, *webnam;
    int atlin, atcol, prelen, pstlen;
    int *cmptr, *chrcnt, *found, i, *lincnt, namlen, *namptr;
    
    /*
    *  Dit telefoonnummer is gekoppeld aan 
    *  .</span>
    */

    ch = 0;
    i = 0;
    namlen = 0;

    /*  Vars  */
    buf    = thrdtab[id].bufs[0];
    chrcnt = &(thrdtab[id].weblkp.chrcnt);
    cmptr  = &(thrdtab[id].weblkp.cmptr);
    found  = &(thrdtab[id].weblkp.found);
    lincnt = &(thrdtab[id].weblkp.lincnt);
    name   = thrdtab[id].phone.name;
    namptr = &(thrdtab[id].weblkp.namptr);
    webnam = thrdtab[id].weblkp.name;
    
    /*  Const  */
    atcol  = lkpsite.atcol;
    atlin  = lkpsite.atlin;
    prelen = lkpsite.prenmlen;
    pstlen = lkpsite.pstnmlen;
    prenam = lkpsite.prename;
    pstnam = lkpsite.pstname;

    while (i < len) {
	ch = buf[i];
	(*chrcnt)++;
	if (ch > 0 && ch < ' ') {
	    /*  Cntrl  */
	    if (ch == 10) {
		/*  New line  */
		(*lincnt)++;
		if (atlin != 0) {
		    /*  Reset character counter  */
		    *chrcnt = 0;
		}
	    }
	    /*  All whitespace to space  */
	    ch = ' ';
	}
	if (*found == 0) {
	    /*  Name not yet found  */
	    if (*lincnt >= atlin && *chrcnt >= atcol) {
		if (prenam[0] != 0) {
		    /*  Pre name based look up  */
		    if (ch == prenam[*cmptr]) {
			(*cmptr)++;
		    } else if (*cmptr != 0) {
			*cmptr = 0;
			/*  Try again  */
			if(ch == prenam[*cmptr])
			    (*cmptr)++;
		    }
		    if (*cmptr >= prelen) {
			/*  Name is next  */
			*found = 1;
		    }
		} else {
		    /*  Look up by line and column alone  */
		    *found = 1;
		}
	    }
	} else {
	    /*
	    *  876543210
	    *  .</span>  Length 8
	    *          ^ Ptr
	    */
	    if (*namptr != 0 || ch != ' ') {
		/*  Ignore leading spaces  */
		webnam[*namptr] = ch;
		(*namptr)++;
		webnam[*namptr] = 0;
	    }
	    if (*namptr > 254) {
		fprintf(flog, "findname(): Name too long\n");
		fflush(flog);
		*found = 2;
		webnam[0] = 0;
		return(-1);
	    }
	    if (*namptr >= pstlen && strncmp(webnam + *namptr - pstlen, \
	      pstnam, pstlen) == 0) {
		/* Finished  */
		namlen = *namptr - pstlen;
		webnam[namlen] = 0;
		*found = 2;
		strncpy(name, webnam, ANLD_NAME_MAX);
		name[ANLD_NAME_MAX] = 0;
		break;
	    }
	}
	i++;
    }

    if (namlen != 0) {
	i = namlen - 1;
	while (i > 0) {
	    if (name[i] == ' ') {
		/*  Remove trailing space  */
		name[i] = 0;
	    } else {
		namlen = i + 1;
		break;
	    }
	    i--;
	}
    }

    return(namlen);
}

/*  Lookup name on web  */
int weblookup(int id)
{
    char *buf, *comnd, *name, *num, *preget, *pstget, *prepost, *pstpost;
    int  *found, inbuf;
    volatile int *cont;

    inbuf = 0;

    /*  Vars  */
    buf   = thrdtab[id].bufs[0];
    cont  = &(thrdtab[id].cont);
    comnd = thrdtab[id].weblkp.comnd;
    found = &(thrdtab[id].weblkp.found);
    name  = thrdtab[id].phone.name;
    num   = thrdtab[id].phone.num;
    /*  Clear  */
    buf[0]   = 0;
    comnd[0] = 0;
    *found   = 0;

    /*  Clear vars used by findname()  */
    thrdtab[id].weblkp.chrcnt  = 0;
    thrdtab[id].weblkp.cmptr   = 0;
    thrdtab[id].weblkp.lincnt  = 0;
    thrdtab[id].weblkp.namptr  = 0;
    thrdtab[id].weblkp.name[0] = 0;

    /*  Const  */
    preget  = lkpsite.preget;
    pstget  = lkpsite.pstget;
    prepost = lkpsite.prepost;
    pstpost = lkpsite.pstpost;

    if (lkpsite.atlin != 0) {
	/*  Line count is used. Start numbering at one  */
	thrdtab[id].weblkp.lincnt = 1;
    }
    if (lkpsite.prenmlen == 0 && lkpsite.atcol == 0 && lkpsite.atlin < 2) {
	/*  Start looking for name right away  */
	*found = 1;
    }

    /*  Command  */
    if (prepost[0] == 0) {
	/*  GET  */
	if (pstget[0] == 0) {
	    snprintf(comnd, 255, "%s '%s%s'", ANLD_CURL_COMND, \
	      preget, num);
	} else {
	    snprintf(comnd, 255, "%s '%s%s%s'", ANLD_CURL_COMND, \
	      preget, num, pstget);
	}
    } else {
	/*  POST  */
	if (pstpost[0] == 0) {
	    snprintf(comnd, 255, "%s -d '%s%s' '%s'", ANLD_CURL_COMND, \
	      prepost, num, preget);
	} else {
	    snprintf(comnd, 255, "%s -d '%s%s%s' '%s'", ANLD_CURL_COMND, \
	      prepost, num, pstpost, preget);
	}
    }

    /*  Open process  */
    thrdtab[id].weblkp.pweb = popen(comnd, "r");
    if (thrdtab[id].weblkp.pweb == NULL) {
	wrtsysl("webookup(): Cant open curl", NULL, 1);
	return(-1);
    }

    if (debflg != 0) {
	/*  Log  */
	wrtolog("Debug: webookup(): num", num);
    }

    /*  Read process  */
    while ((inbuf = fread(buf, 1, 4095, thrdtab[id].weblkp.pweb)) > 0) {
	buf[inbuf] = 0;
	if (*cont == 0) {
	    /*  Premature end  */
	    pclose(thrdtab[id].weblkp.pweb);
	    thrdtab[id].weblkp.pweb = NULL;
	    wrtsysl("webookup(): Terminated", NULL, 1);
	    return(-1);
	}
	if (*found < 2) {
	    //findname(buf, inbuf, &(site.nam), name);
	    findname(id, inbuf);
	}
    }
    if (name[0] != 0) {
	/*  Found name  */
	wrtsysl("weblookup", name, 0);
    }

    pclose(thrdtab[id].weblkp.pweb);
    thrdtab[id].weblkp.pweb = NULL;

    return(0);
}


/*  Main  */
int main(int argc, char **argv)
{
    int confd, i, opt;
    char *home, *line, *lognam, *logstr, *pidnam;
    FILE *confil;
    /*  PID  */
    pid_t pid;
    DIR  *piddir;
    FILE *pidfil;
    /*  Sockets  */
    int port;
    socklen_t adrlen;
    struct sockaddr_in6 addr, caddr;
    struct in6_addr *remadr;
    /*  Select stuff  */
    int maxfd, ret;
    fd_set readfds;
    struct timeval tv;
    /*  User  */
    gid_t gid;
    uid_t uid; 
    struct group *group;
    struct passwd *passwd;

    #ifdef AGID_DAEMON
     /*  Demonize by fork and exit  */
     pid = 0;
     /*  Fork  */
     if ((pid = fork()) < 0) {
	 wrtsysl("main(): fork()", strerror(errno), 1);
	 exit(1);
     }
     if (pid != 0) {
	 /*  Parent  */
	 exit(0);
     }
     /*  Child keeps running  */
    #endif

    aclflg = 0;
    blkent = 0;		/*  Number of blocklist entries  */
    gid    = 0;		/*  Group ID  */
    i      = 0;
    maxfd  = 0;
    opt    = 0;
    pid    = 0;		/*  PID  */
    ret    = 0;
    blkent = 0;		/*  Block list number of entries  */
    blksiz = 0;		/*  Block list size  */
    disweb = 0;		/*  Disable web lookup when not zero  */
    numacl = 0;		/*  Nr of entries in ACL  */
    uid    = 0;		/*  User ID  */
    acltab  = NULL;	/*  ACL Table  */
    blkdb   = NULL;	/*  Block list database  */
    confil  = NULL;	/*  Conf file  */
    flog    = NULL;	/*  Log file  */
    home    = NULL;	/*  Home directory  */
    group   = NULL;
    lognam  = NULL;	/*  Log file name  */
    passwd  = NULL;
    piddir  = NULL;	/*  PID file directory  */
    pidfil  = NULL;	/*  PID file  */
    pidnam  = NULL;	/*  PID file name  */
    remadr  = NULL;	/*  Remote IP address  */
    thrdtab = NULL;	/*  Thread table  */
    memset(confdir,  0, 256);	/*  Alternate conf dir  */
    memset(genbuf,   0, 4096);	/*  General purpose buffer */
    memset(glogs,    0, 256);	/*  Global string for log  */
    memset(ipstr,    0, 48);	/*  Text version of IP address  */
    memset(&addr,    0, sizeof(struct sockaddr_in6));
    memset(&caddr,   0, sizeof(struct sockaddr_in6));
    memset(&areas,   0, sizeof(struct phndb));
    memset(&numbers, 0, sizeof(struct phndb));
    memset(&regions, 0, sizeof(struct phndb));
    memset(&tv,      0, sizeof(struct timeval));
    memset(&lkpsite, 0, sizeof(struct urlprps));
    FD_ZERO(&readfds);

    /*  Defaults  */
    adrlen   = sizeof(struct sockaddr_in6);
    contprog = 1;
    confd    = -1;
    lsfd     = -1;
    numthrds = ANLD_DFLT_THRDS;
    port     = ANLD_DFLT_PORT;

    /*  Buffers  */
    line   = genbuf;
    logstr = glogs;

    /*  mbstowcs() and wcstombs() need this  */
    setlocale(LC_ALL, "C.UTF-8");

    /*  Get UID  */
    uid = getuid();

    while ((opt = getopt(argc, argv, "ac:dhn:p:u:v")) != -1) {
	switch (opt) {
	    case 'a':
		aclflg = 1;
		break;
	    case 'c':
		strncpy(confdir, optarg, 255);
		break;
	    case 'd':
		debflg = 1;
		break;
	    case 'h':
		hlp();
		exit(0);
	    case 'n':
		numthrds = atoi(optarg);
		if (numthrds < 2 || numthrds > 1024) {
		    numthrds = ANLD_DFLT_THRDS;
		    wrtsysl("main(): Number of threads should be 2 - 1024", \
		      NULL, 1);
		    wrtsysl("main(): Using default", NULL, 1);
		}
		break;
	    case 'p':
		port = atoi(optarg);
		if (port < 1024 || port > 65535) {
		    port = ANLD_DFLT_PORT;
		    wrtsysl("main(): Port should be 1024 - 65535", NULL, 1);
		    wrtsysl("main(): Using default", NULL, 1);
		}
		break;
	    case 'u':
		/*  Try numeric first  */
		uid = (uid_t) atoi(optarg);
		if (uid < 1 || uid > 65535) {
		    /*  Try user name  */
		    passwd = getpwnam(optarg);
		    if (passwd == NULL) {
			wrtsysl("main(): User not found", NULL, 1);
			exit(1);
		    } else {
			uid = passwd->pw_uid;
			if (gid == 0) {
			    /*  Set group if not set  */
			    group = getgrnam(optarg);
			    if (group == NULL) {
				wrtsysl("main(): Group not found", NULL, 1);
				exit(1);
			    }
			    gid = group->gr_gid;
			}
		    }
		} else if (gid == 0) {
		    gid = (gid_t) uid;
		}
		break;
	    case 'v':
		prtvers();
		exit(0);
	    default:
		hlp();
		exit(1);
	}
    }

    if (confdir[0] != 0) {
	/*  Use alternate config dir  */
	if (chdir(confdir) != 0) {
	    wrtsysl("main(): Cant change dir", strerror(errno), 1);
	    exit(1);
	}
    }

    /*  Read config file if any  */
    if (confdir[0] == 0)
	confil = fopen(ANLD_CONFIG_FILE, "r");
    else
	confil = fopen(ANLD_CONF_FIL, "r");
    if (confil != NULL) {
	while (fgets(line, 4096, confil)) {
	    if(line[0] == 0)
		continue;
	    if (strncasecmp(line, "weblookup=", 10) == 0) {
		if (line[10] == '0' || \
		  strncasecmp(line + 10, "no", 2) == 0 || \
		  strncasecmp(line + 10, "off", 3) == 0 || \
		  strncasecmp(line + 10, "false", 5) == 0) {
		    /*  Disable web lookups  */
		    disweb = 1;
		} else if (line[10] == '1' || \
		  strncasecmp(line + 10, "yes", 3) == 0 || \
		  strncasecmp(line + 10, "on", 2) == 0 || \
		  strncasecmp(line + 10, "true", 4) == 0) {
		    /*  Enable web lookups (default)  */
		    disweb = 0;
		}
  	    }
	}
	fclose(confil);
	if (debflg != 0) {
	    if(disweb != 0)
		wrtsysl("main: Web lookups disabled", NULL, 0);
	}
    }

    /*  Get Pid  */
    pid = getpid();
    #ifdef AGID_DAEMON
     /*  Close  */
     fclose(stdin);
     fclose(stderr);
     fclose(stdout);
     /*  Become session leader  */
     if (setsid() < 0) {
	 wrtsysl("main(): setsid()", strerror(errno), 1);
	 exit(1);
     }
    #else
     /*  Become program group leader  */
     if (setpgid(pid, pid) != 0) {
	 wrtsysl("main(): setpgid()", strerror(errno), 1);
	 exit(1);
     }
    #endif

    /*  Write pid file now if pid dir doesn't exist  */
    if (confdir[0] == 0) {
	pidnam = ANLD_PID_FILE;
	if ((piddir = opendir(ANLD_PID_PATH)) == NULL) {
	    /*
	    *  Directory doesn't exist;
	    *  Write as root in alternate pid dir
	    */
	    pidnam = ANLD_PID_FILE2;
	    if ((pidfil = fopen(pidnam, "w")) == NULL) {
		wrtsysl("main(): Cant open", ANLD_PID_FIL, 1);
		exit(1);
	    } else {
		fprintf(pidfil, "%d\n", (int) pid);
		fclose(pidfil);
	    }
	} else {
	    closedir(piddir);
	}
    } else {
	pidnam = ANLD_PID_FIL;
    }

    /*  Uid and Gid  */
    if (gid != 0 && setgid(gid) < 0) {
	wrtsysl("main(): Cant set GID", strerror(errno), 1);
    }
    if (uid != 0 && setuid(uid) < 0) {
	wrtsysl("main(): Cant set UID", strerror(errno), 1);
    }

    /*  Set home dir  */
    if (confdir[0] == 0) {
	/*  Use default conf dir  */
	if (uid != 0) {
	    /*  Set home directory  */
	    if (passwd == NULL)
		passwd = getpwuid(uid);
	    if (passwd != NULL) {
		home = passwd->pw_dir;
		chdir(home);
		setenv("HOME", home, 1);
	    }
	}
	/*  Location of .curlrc  */
	setenv("CURL_HOME", ANLD_CONF_PATH, 1);
    } else {
	/*  Use alternate config dir  */
	setenv("HOME", confdir, 1);
    }

    /*  Log file  */
    if (confdir[0] == 0)
	lognam = ANLD_LOG_FILE;
    else
	lognam = ANLD_LOG_FIL;
    if ((flog = fopen(lognam, "a")) == NULL) {
	wrtsysl("main(): Cant open", lognam, 1);
	exit(1);
    }

    /*  Log start to syslog  */
    wrtsysl("main: Version", version, 0);
    if(debflg != 0)
	wrtsysl("main: Debug on", NULL, 0);
    if (confdir[0] != 0)
	wrtsysl("main: Config dir", confdir, 0);

    /*  Log version  */
    wrtolog("main(): Version", version);

    /* Log pid  */
    snprintf(logstr, 255, "main(): Pid: %d", (int) pid);
    wrtolog(logstr, NULL);
    /*  If alternate pid file isn't used  */
    if (pidnam != ANLD_PID_FILE2) {
	/*  Alternate pid file isn't used  */
	if ((pidfil = fopen(pidnam, "w")) == NULL) {
	    snprintf(logstr, 255, "main(): Cant open %s", ANLD_PID_FIL);
	    wrtolog(logstr, NULL);
	    if(flog != NULL)
		fclose(flog);
	    exit(1);
	}
	fprintf(pidfil, "%d\n", pid);
	fclose(pidfil);
    }

    /*  Initialize tables  */
    if (init() != 0) {
	quit();
	exit(1);
    }

    /*  Open listening socket  */
    if ((lsfd = openlsock(&addr, (uint16_t) port, 2)) < 0) {
	quit();
	exit(1);
    }

    /*  Stop on TERM  */
    signal(SIGTERM, stop);
    /*  Without catching SIGPIPE, a write() won't detect a closed socket  */
    signal(SIGPIPE, SIG_IGN);
    /*  Set signal for log file rotation  */
    signal(SIGUSR1, rotlog);

    wrtolog("namelookup.agid started", NULL);

    /*  Main loop  */
    while (contprog != 0) {
	/*  (Re) Initialise select()  */
	FD_ZERO(&readfds);
	maxfd = lsfd;
	FD_SET(lsfd, &readfds);
	maxfd++;
	tv.tv_usec = 0;
	tv.tv_sec  = ANLD_CHK_TRD_TIM;
	ret = select(maxfd, &readfds, NULL, NULL, &tv);
	if (ret < 0) {
	    if (errno == EINTR || errno == EAGAIN || errno == EWOULDBLOCK) {
		continue;
	    } else {
		wrtsysl("main(): Error select", strerror(errno), 1);
		contprog = 0;
		break;
	    }
	}
	/*
	*  A thread may have finished but didn't terminate just yet.
	*  Do a join to wait for actual termination.
	*/
	i = 0;
	while (i < numthrds) {
	    if (thrdtab[i].cont == 0 && thrdtab[i].sfd < 0 && \
	      thrdtab[i].thrdid != 0) {
		/*  Finished thread  */
		if (pthread_join(thrdtab[i].thrdid, NULL) != 0) {
		    snprintf(logstr, 255, \
		      "main(): pthread_join() for thread nr: %d failed\n", i);
		    wrtsysl(logstr, strerror(errno), 1);
		} else {
		    thrdtab[i].thrdid = 0;
		}
	    }
	    i++;
	}
	/*  Check listening socket  */
	if (ret != 0 && FD_ISSET(lsfd, &readfds)) {
	    memset(&caddr, 0, sizeof(struct sockaddr_in6));
	    if ((confd = (accept(lsfd, (struct sockaddr *) &caddr, \
	      &adrlen))) < 0) {
		if (errno != EINTR && errno != EAGAIN && errno != EWOULDBLOCK) {
		    wrtsysl("main(): Error accept", strerror(errno), 1);
		    contprog = 0;
		    break;
		}
            } else {
		remadr = &(caddr.sin6_addr);
		if (inet_ntop(AF_INET6, remadr, ipstr, 47) == NULL) {
		    wrtsysl("main(): inet_ntop() failed", \
		      strerror(errno), 1);
		} else {
		    if (strncmp(ipstr, "::ffff:", 7) == 0) {
			wrtsysl("main: connect from", ipstr + 7, 0);
		    } else {
			wrtsysl("main: connect from", ipstr, 0);
		    }
		}
		if (acltab == NULL || chkacl(remadr) == 0) {
		    /*  Assign to thread  */
		    i = 0;
		    while (i < numthrds) {
			if (thrdtab[i].cont == 0 && thrdtab[i].sfd < 0 && \
			  thrdtab[i].thrdid == 0) {
			    /*  Free thread entry  */
			    if (debflg != 0) {
				snprintf(logstr, 255, \
				"Debug: main(): new thread nr %d", i);
				wrtolog(logstr, NULL);
			    }
			    thrdtab[i].cont = 1;
			    thrdtab[i].sfd  = confd;
			    pthread_create(&(thrdtab[i].thrdid), NULL, \
			      proclsock, (void *) &(thrdtab[i].id));
			    break;
			}
			i++;
		    }
		    if (i >= numthrds) {
			wrtsysl("main(): All threads are used", NULL, 1);
			close(confd);
		    }
		} else {
		    wrtsysl("main(): Forbidden by ACL", NULL, 1);
		    close(confd);
		}
	    }
	}
    }

    /*  Stop all threads  */
    i = 0;
    while (i < numthrds) {
	if (thrdtab[i].thrdid != 0 && thrdtab[i].cont != 0) {
	    thrdtab[i].cont = 0;
	}
	i++;
    }
    /*
    *  Wait before kill.
    *  If this time is too short, increase in common.h
    */
    usleep(ANLD_KIL_THRD_TIM);
    /*  Kill all threads  */
    i = 0;
    while (i < numthrds) {
	if (thrdtab[i].thrdid != 0 && thrdtab[i].sfd > -1) {
	    if (pthread_cancel(thrdtab[i].thrdid) != 0) {
		snprintf(logstr, 255, \
		  "main(): pthread_cancel() for thread nr: %d failed\n", i);
		wrtsysl(logstr, strerror(errno), 1);
	    }
	}
	i++;
    }
    if (debflg != 0) {
	/*  Join all threads  */
	i = 0;
	while (i < numthrds) {
	    if (thrdtab[i].thrdid != 0) {
		if (pthread_join(thrdtab[i].thrdid, NULL) != 0) {
		    snprintf(logstr, 255, \
		      "main(): pthread_join() for thread nr: %d failed\n", i);
		    wrtsysl(logstr, strerror(errno), 1);
		}
	    }
	    i++;
	}
    }

    signal(SIGUSR1, SIG_DFL);
    signal(SIGPIPE, SIG_DFL);
    signal(SIGTERM, SIG_DFL);

    if(flog != NULL)
	wrtolog("namelookup.agid stopped", NULL);
    quit();

    if(unlink(pidnam) < 0)
	wrtsysl("main(): Cant remove pidfile", strerror(errno), 1);

    wrtsysl("main: Stopped", NULL, 0);

    return(0);
}

